#!/usr/bin/python


#
# RSVP: Rapid Serial Visual Presentation
#
# A text reader designed for reading etexts on linux-based PDAs.
# Displays one word at a time, quickly, in a large font.
# At least, that's the original idea.  More display modes have
# been added for better readability.
#
# Concepts based on GNU sview and Panamedia PanaRead.
#
# Original author: Scott Scriven <gutenflash@xyzz.org>
#
# This program is distributed under the GNU Public License.
#


#
# TODO: 
#  - make toolbar app thing work (setting gdk atoms, etc)
#  - add delays for long words.
#  - html input (and "Open URL...")
#  - configurable punctuation maps (user defines what punctuation
#    counts as "short", "medium", and "long")
#  - smallfont text map? (birds eye view)
#  - bookmarks?
#  + optimize memory use!!!
#    * minimize use when entire text is loaded
#    - load only part of the text at a time
#  * Chapter chooser window
#  * percent meter
#  * percent chooser?  (menu worked, but slider is better)
#  * automatic chapter detection
#    (implemented by searching for "^CHAPTER" and similar
#  * chapter selection
#  * 3-label bigtext widget?
#    (used a 3-word single label instead; easier to read!)
#    (multiword mode is even better)
#  * Detect and correct cases where the text doesn't fit!
#    (implemented as multiword mode)
#  * Make multiword mode's timing dependent on what is visible
#    rather than what is being buffered in the next word set.
#    (painful hairy messy implementation probably required?)
#  * display chapters while reading.
#  * make fonts work again!
#

program_name = "gutenflash"
program_version = "0.8.4"


# Importing gtk takes too long, especially on an ipaq.
#print "Loading gtk...."
import pygtk
pygtk.require("2.0")
import gtk
#print "Loading gtk....  done."

from tktk import Run, Quit, runqueue, ButtonChoiceWindow, ListBoxChoiceWindow

class plaintext:
  """Loads regular text files into the required native data structure."""
  def __init__(self, filename=None):
    self.filename = filename
    self.text = ""
    self.lines = []
    self.wpl = 64  # words per line
    self._line = -1
    self.words = [] # words in current line
    self.types = []
    self.chapters = []
    self.length = 0

    self.chaptertexts = [
      "CONTENTS",
      "INDEX",
      "FOREWORD",
      "PROLOGUE",
      "CHAPTER ", "Chapter",
      "PART",
      "BOOK",
      "SECTION ", "Section",
      "EPILOGUE",
      "GLOSSARY",
      "APPENDIX",
      ]

    self.abbreviations = [
      "Dr.", "Mr.", "Mrs.", "Ms.",
      #"A.", "B.", "C.", "D.", "E.", "F.", "G.", "H.",       "J.", "K.",
      #"L.", "M.", "N.", "O.", "P.", "Q.", "R.", "S.", "T.", "U.", "V.",
      #"W.", "X.", "Y.", "Z.",
      ]

    self.win = gtk.Window()
    self.vbox = gtk.VBox()
    self.win.add(self.vbox)
    self.label = gtk.Label()
    self.progress = gtk.ProgressBar()
    self.vbox.add(self.label)
    self.vbox.add(self.progress)

  def load_line(self, line):
    import string
    
    self._line = line
    self.words = string.split(self.lines[self._line], " ")

  def word(self, num):
    if self.length <= 0:
      return ""
    
    line = num / self.wpl
    offset = num % self.wpl
    
    #print "getword(%s): line(%s/%s)" % (num, line, len(self.lines))
    if line != self._line:
      self.load_line(line)

    #try:
    word = self.words[offset]
    #except:
    #  print "word %s/%s in: %s" % (offset, len(self.words), self.words)
    #  raise "hell"
    
    #print "Word(%s): '%s'" % (num, word)

    return word

  def type(self, num):
    if self.length <= 0:
      return " "
    return self.types[num]

  def len(self):
    if self.length <= 0:
      return 1
    else:
      return self.length

  def add_word(self, w, t):
    """
    w must be non-space, non-null
    t must be 1 letter
    """
    # TODO: store current line as a list instead of a string,
    #       for easier popping and less regenerating text, and
    #       join it together when it reaches appropriate length...

    if t == "c":
      self.chapters.append((w, self.length))
      w = "."

    #print "Word(%s): %s (%s)" % (self.length, w, t)

    self.types.append(t)

    if (self.length % self.wpl == 0):  #or (len(self.lines) == 0):
      self.lines.append(w)
    else:
      #self.lines[-1] = "%s %s" % (self.lines[-1], w)
      self.lines[-1] = self.lines[-1] + " " + w

    self.length = self.length + 1

  def pop_word(self):
    import string
    
    if self.length <= 0:
      return

    #print "pop_word() Length(%s), lines(%s)" % (self.length, len(self.lines))

    if len(self.lines[-1]) == 0:
      #print "Popping line(a): %s" % self.lines[-1]
      self.lines.pop()
      
    elif self.length % self.wpl == 1:
      #print "Popping line(b): %s" % self.lines[-1]
      self.lines.pop()
    #elif self.length % self.wpl == 0:
    #  pass
    else:
      line = self.lines[-1]
      offset = string.rfind(line, " ")
      self.lines[-1] = line[:offset]
      
    self.types.pop()

    self.length = self.length - 1

  def make_test_words(self):
    words = [
      "Press", ">", "to", "play", "back", "text.", ["."],
      "The", "[]", "button", "stops", "playback.", ["."],
      "Press", ">[]", [","], "to", "step", "one", "frame", "at",
      "a", "time.", ["."],
      "Use", "|<<", [","],"and", ">>|", [","], "to", "skip", "back", "or",
      "forward", "by", "one", "paragraph,", [","], "and", "|<", [","], "to",
      "skip", "back", "by", "one", "sentence.", ["\n"],
      "Use", "the", "top", "slider", "below", "to", "navigate",
      "quickly", "through", "the", "text.", ["."],
      "The", "bottom", "slider", "below", "controls", "the", "playback",
      "speed.", ["\n"],
      ]

    for w in words:
      if type(w) == type("hi"):
        self.add_word(w, " ")
      else:
        self.add_word(".", w[0])

  def load(self, filename=None, viewer_win=None, enter_is_pbreak=0):
    import string
    import sys

    if filename:
      self.filename = filename

    try:
      if len(self.filename) > 3  and  self.filename[-3:] == ".gz":
        import gzip
        fp = gzip.open(self.filename, "rb")
      else:
        if self.filename == "-":
          fp = sys.stdin
        else:
          fp = open(self.filename, "rb")
    except Exception, msg:
      print "Couldn't load %s: %s" % (self.filename, msg)
      return

    print "Loading %s..." % (self.filename)

    # Set up progress bar.
    self.label.set_text("Loading %s..." % (self.filename))
    self.progress.set_fraction(0.0)
    #self.progress.set_usize(100, 15)
    self.progress.set_size_request(100, 15)
    self.win.set_transient_for(viewer_win)
    self.win.set_modal(1)
    self.win.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
    self.win.show_all()
    runqueue()

    self.types = []
    prev = None

    current_line = 0
    lines = fp.readlines()
    self.text = ""
    numlines = len(lines)
    update_freq = numlines / 200
    if update_freq < 1: update_freq = 1
    #print "freq: %s" % update_freq

    for line in lines:
      # Update the progress bar
      current_line = current_line + 1
      if current_line % update_freq == 0:
        self.progress.set_fraction(float(current_line) / numlines)
        runqueue()
      #print "Line: %s" % (line)

      # Check for chapter headings.
      is_chapter = 0
      stripline = string.strip(line)
      for c in self.chaptertexts:
        if stripline[:len(c)] == c:
          is_chapter = 1
      if is_chapter:
        #print "Chapter: %s (%s)" % (line, len(self.wordlist))
        self.add_word(stripline, "c")

        ws = string.split(stripline)
        for w in ws:
          self.add_word(w, " ")
        self.add_word(".", "\n")
        continue

      # Paragraph break signified by blank lines or indentation
      is_paragraph = 0
      if stripline == "":
        # Ignore leading whitespace
        if self.length == 0:
          continue
        # Compensate for texts with blank lines in the middle of
        # a paragraph.  Ignore paragraph breaks in the middle of sentences.
        if prev == " ":
          continue

        is_paragraph = 1

      # paragraph declared by indentation
      if line[:3] == "   "  or  line[0] == "\t":

        # Compensate for indented blocks of text
        if prev != " ":
          is_paragraph = 1
        

      if is_paragraph:
        # Avoid two different pauses in a row
        # Paragraphs and override regular punctuation
        if prev in [".", ",", "-", ":"]: # prev != ""
          self.pop_word()

        # Chapters override paragraphs.
        # Avoid repeating paragraph breaks.
        if prev not in ["c", "\n"]:
          self.add_word(".", "\n")
          prev = "\n"

      # Parse the words...
      words = string.split(line)
      for w in words:
        #w = string.strip(w)
        #if w == "": continue

        # Re-join hyphen-ated wo-rds
        # FIXME: Broken right now...
        if prev == "-":
          #w2 = self.word(self.length - 1)
          #w2 = w2[:-1]
          #self.pop_word()
          #w = "%s%s" % (w2, w)
          self.add_word(w, " ")
        else: # Regular word
          # FIXME: make double-hyphenation check faster!
          # Check for internal double-hyphenation  "better--and"
          # ("-" check is to eliminate need for "--" check in most words)
          #if "-" in w  and  (string.find(w, "--") > 0):
          #  i1 = string.find(w, "--")
          #  i2 = string.rfind(w, "--") + 2
          #  w1 = w[:i1]
          #  w2 = w[i1:i2]
          #  w3 = w[i2:]
          #  self.add_word(w1, " ")
          #  self.add_word(w2, " ")
          #  self.add_word(w3, " ")
          #  w = w3
	  #  if w == "":  # "sometimes I think--"
	  #    continue
          # Otherwise, regular word
          #else:
          #  self.add_word(w, " ")
          self.add_word(w, " ")
          
        prev = " "

        # Handle punctuation.
        if w[-1] in '.!?"':
          if w not in self.abbreviations:
            self.add_word(".", ".")
            prev = "."
        elif w[-1] in ",;":
          self.add_word(".", ",")
          prev = ","
        elif w[-1] in ":\'\")]}>":
          self.add_word(".", ":")
          prev = ":"
        # Setup for de-hyphenation of words.
        elif w[-1] == "-":
          if len(w) > 1  and  w[-2] in "-":  # "sometimes I think--"
            self.add_word(".", ":")
            prev = ":"
          elif len(w) > 1: #  "then he per- \n formed the operation"
            prev = "-"
          else: # just a dash
            pass

      # Handle paragraph breaks denoted by a \n
      if enter_is_pbreak:
        # Avoid two different pauses in a row
        # Paragraphs and override regular punctuation
        if prev in [".", ",", "-", ":"]: # prev != ""
          self.pop_word()

        # Chapters override paragraphs.
        # Avoid repeating paragraph breaks.
        if prev not in ["c", "\n"]:
          self.add_word(".", "\n")
          prev = "\n"

    self.types = string.join(self.types, "")

    fp.close()
    self.win.hide()
    self.win.unrealize()
    runqueue()
    
    print "Loading %s...  Done." % (self.filename)

  def __del__(self, *args):
    #self.win.hide()
    #runqueue()
    pass

######################################################################
# Not used.
# Instead, a simple 2-tuple is used.
# It uses less than half as much memory!
# (and this is intended for use on an ipaq, with limited resources)
#
#class word:
#
#  def __init__(self, value, style=""):
#    self.value = value
#    # Possible styles:
#    # - ,
#    # - :
#    # - .
#    # - chapter
#    # - \n
#    self.style = style
#    if not value:
#      self.value = ""

def set_gdk_win_prop(window, property, type,
                     format, mode, data):
  import _gtk, GDK
  atom = _gtk.gdk_atom_intern(property, 1)
  if not int(atom): return 0
  int_atom = _gtk.gdk_atom_intern(type, 1)
  if not int(int_atom): return 0
  if mode=="REPLACE":gdk_mode = GDK.PROP_MODE_REPLACE
  elif mode=="APPEND":gdk_mode = GDK.PROP_MODE_APPEND
  elif mode=="PREPEND":gdk_mode = GDK.PROP_MODE_PREPEND
  try:
    window.property_change(atom, int_atom, format, gdk_mode, data)
    return 1
  except: return 0

class viewer:
  def __init__(self, cfg):
    self.cfg = cfg
    self.book = None
    self.win = gtk.Window()
    self.win.connect("destroy", self._destroy)
    self.vbox = gtk.VBox()
    self.win.add(self.vbox)

    self.make_menus()

    self.make_textbox()
    
    self.wpm = self.cfg.words_per_minute

    self.direction = 1

    self.make_speed_bar()
    self.make_play_buttons()

    self.filename = ""
    self.shortfilename = ""

    self._offset = 0
    self._pausecount = 0

    self._docked = 0
    if self.cfg.toolbar_app:
      self.set_docked(1, show=0)
    self.win.set_title("%s %s" % (program_name, program_version))
    self.win.show_all()
    runqueue()

  def nop(self, *args):
    pass

  def _toggle_docked(self, *args):
    self.set_docked(not self._docked)

  def set_docked(self, dock, show=1):
    import _gtk
    
    self.win.realize()
    if show:
      self.win.hide()
    if dock:
      t = [_gtk.gdk_atom_intern("_NET_WM_WINDOW_TYPE_TOOLBAR")]
    else:
      t = []
    set_gdk_win_prop(self.win.get_window(), "_NET_WM_WINDOW_TYPE",
                     "ATOM", 32, "REPLACE", t)
    if show:
      # FIXME: can't resize when docking
      #self.textbox.set_usize(self.cfg.view_wid,
      #                       self.cfg.view_hgt)
      self.win.show_all()
    self._docked = dock
    self._displayed = 0

  def load_plaintext(self, infile):
    import string
    self.filename = infile
    index = string.rfind(infile, "/") + 1
    self.shortfilename = infile[index:]
    
    self.book = plaintext()
    self.book.load(infile, self.win,
                   self.cfg.enter_is_paragraph_break)
    #self.update_chapter_menu()
    self.win.set_title("%s: %s" % (program_name, self.shortfilename))
    #import sys
    #sys.exit(0)

  def make_menus(self):
    self.ag = gtk.AccelGroup()
    self.win.add_accel_group(self.ag)
    self.itemf = gtk.ItemFactory(gtk.MenuBar, "<main>", self.ag)
    self.itemf.create_items([
      ("/_File",       None, None, 0, "<Branch>"),
      ("/File/_Open",  "<Control>O", self._file_open, 0, ""),
      ("/File/Calculate _Length",  "<Control>L", self._calc_length, 0, ""),
      ("/File/Calculate _Remaining",  "<Control>R", self._calc_remaining, 0, ""),
      #("/File/Toggle Dock" ,  None, self._toggle_docked, 0, ""),
      ("/File/_Quit",  "Q", self._destroy, 0, ""),
      #("/_Bookmarks",   None, None, 0, "<Branch>"),
      #("/_Chapters",   None, None, 0, "<Branch>"),
      ("/_Jump",       None, None, 0, "<Branch>"),
      ("/Jump/To _Chapter...",  "C", self._select_chapter, 0, ""),
      ("/Jump/Next Chapter",  None, self._next_chapter, 0, ""),
      ("/Jump/Previous Chapter",  None, self._prev_chapter, 0, ""),
      ("/Jump/Start",  None, self._percent_jump, 0, ""),
      ("/Jump/End",    None, self._percent_jump, 100, ""),
      ("/_Mode",       None, None, 0, "<Branch>"),
      ("/Mode/One Word",    None, self._set_mode, 0, ""),
      ("/Mode/Three Words", None, self._set_mode, 1, ""),
      ("/Mode/Multiword",   None, self._set_mode, 2, ""),
      ("/_Play",       None, None, 0, "<Branch>"),
      ("/Play/_Play",  "P", self._play_button, 0, ""),
      ("/Play/St_ep",  "E", self._step_button, 0, ""),
      ("/Play/_Stop",  "S", self._stop_button, 0, ""),
      ("/Play/Back (sentence)", "R", self._rplay_button, 0, ""),
      ("/Play/_Back (paragraph)", "B", self._rskip_button, 0, ""),
      ("/Play/_Forward",  "F", self._skip_button, 0, ""),
      ("/_About",       None, None, 0, "<LastBranch>"),
      ("/About/About %s" % (program_name), None, self.about_screen, 0, ""),
      ])
    self.menubar = self.itemf.get_widget("<main>")
    #self.chapter_menu = self.itemf.get_widget("/Chapters")
    self.vbox.pack_start(self.menubar, 0, 0)

    # FIXME: these are being ignored?
    foo = gtk.ItemFactory(gtk.MenuBar, "<keydefs>", self.ag)
    foo.create_items([
      ("/foo",       None, None, 0, "<Branch>"),
      ("/foo/rskip", "B", self._rskip_button, 0, ""),
      ("/foo/rplay", "R", self._rplay_button, 0, ""),
      ("/foo/stop",  "S", self._stop_button, 0, ""),
      ("/foo/step",  "E", self._step_button, 0, ""),
      ("/foo/play",  "P", self._play_button, 0, ""),
      ("/foo/skip",  "F", self._skip_button, 0, ""),
      ])
    
  def update_chapter_menu(self):
    for c in self.book.chapters:
      #print "Chapter(%s): %s" % (c[1], c[0])
      mi = gtk.MenuItem(c[0])
      mi.connect("activate", self._chapter_jump, c[1])
      self.chapter_menu.append(mi)

  def make_textbox(self):
    import pango
    self._last_text = ""
    self.textbox = gtk.Label()
    self.textbox.set_size_request(self.cfg.view_wid, self.cfg.view_hgt)

    self.font = pango.FontDescription(self.cfg.bigfont)
    self.textbox.modify_font(self.font)

    self.timer = None

    self.vbox.pack_start(self.textbox, 1, 1)

  def make_play_buttons(self):
    self.button_box = gtk.HBox()
    self.rskip_button = gtk.Button("|<<")
    self.button_box.add(self.rskip_button)
    self.rplay_button = gtk.Button("|<")
    self.button_box.add(self.rplay_button)
    self.stop_button = gtk.Button("[]")
    self.button_box.add(self.stop_button)
    self.step_button = gtk.Button(">[]")
    self.button_box.add(self.step_button)
    self.play_button = gtk.Button(">")
    self.button_box.add(self.play_button)
    self.skip_button = gtk.Button(">>|")
    self.button_box.add(self.skip_button)
    self.vbox.pack_start(self.button_box, 0, 1)

    self.stop_button.connect("clicked", self._stop_button)
    self.step_button.connect("clicked", self._step_button)
    self.play_button.connect("clicked", self._play_button)
    self.rplay_button.connect("clicked", self._rplay_button)
    self.skip_button.connect("clicked", self._skip_button)
    self.rskip_button.connect("clicked", self._rskip_button)

  def make_speed_bar(self):
    self._ignore_percent_events = 0
    self.percent_adjustment = gtk.Adjustment(0.0, 0.0, 100.0,
                                          0.1, 5.0, 0.0)
    self.percent_adjustment.set_value(0.0)
    self.percent_adjustment.connect("value-changed", self._handle_percentbar)
    self.percentbar = gtk.HScale(self.percent_adjustment)
    self.percentbar.set_draw_value(1)
    self.percentbar.set_digits(1)
    self.vbox.pack_start(self.percentbar, 0, 1)
    
    #self.speed_hbox = gtk.HBox()
    #self.vbox.pack_start(self.speed_hbox, 0, 1)
    self.speed_adjustment = gtk.Adjustment(self.wpm, 20.0, 2000.0,
                                          5.0, 50.0, 0.0)
    self.speed_adjustment.set_value(self.wpm)
    self.speed_adjustment.connect("value-changed", self._handle_speedbar)
    self.speedbar = gtk.HScale(self.speed_adjustment)
    self.speedbar.set_draw_value(1)
    self.speedbar.set_digits(0)
    #self.speedbar.set_value_pos(POS_RIGHT)
    self.vbox.pack_start(self.speedbar, 0, 1)

    #self.vbox.pack_start(self.speedbar, 0, 1)
    #self.percent_label = gtk.Label("0%")
    #self.speed_hbox.pack_start(self.speedbar, 1, 1)
    #self.speed_hbox.pack_start(self.percent_label, 0, 0)

  def run(self):
    self.win.show_all()
    self.show_word(flush=1)
    #self.win.set_title("%s: %s" % (program_name, self.shortfilename))
    gtk.mainloop()

  def about_screen(self, *args):
    text = \
"""%s V%s

This is a text reader based on the concept of RSVP.  Rapid Serial Visual Presentation.  The idea is to display one word at a time, very quickly and in a large font.

At least, that's the original idea.  More display modes have been added for better readability.

Concepts based on GNU sview and Panamedia PanaRead.

Original author: Scott Scriven <gutenflash@toykeeper.net>

This program is distributed under the GNU Public License.
""" % (program_name, program_version)
    ButtonChoiceWindow()("About...", text, ["Close"], long=1, transient=self.win)

  def _about_destroy(self, *args):
    gtk.mainquit()
    self.about_win.hide()
    self.about_win.unrealize()
    del self.about_win

  def _destroy(self, *args):
    self.win.hide()
    self.win.unrealize()
    del self.win
    gtk.mainquit()
    for key in self.__dict__.keys():
      del self.__dict__[key]

  def _chapter_jump(self, menuitem, data):
    self._offset = data
    self.show_word(flush=1)
    self._displayed = 0

  def _percent_jump(self, data, menuitem):
    self._offset = int(data * self.book.length / 100.0)
    self.show_word(flush=1)

  def _file_open(self, *args):
    filename = self.file_selector("Open...", self.cfg.load_dir)
    if filename:
      v = viewer(self.cfg)
      self._destroy()
      v.load_plaintext(filename)
      v.run()
    
  def file_selector(self, title, filename):
    self.file_selector_cb_button = None
    
    fs = gtk.FileSelection(title)
    if filename:
      fs.set_filename(filename)
    fs.ok_button.connect("clicked", self.file_selector_ok_cb)
    fs.cancel_button.connect("clicked", self.file_selector_cancel_cb)
    fs.show_all()
    gtk.mainloop()

    if self.file_selector_cb_button == "ok":
      filename = fs.get_filename()
      fs.hide()
      fs.unrealize()
      return filename
    else:
      fs.hide()
      fs.unrealize()
      return None

  def file_selector_ok_cb(self, *args):
    #print "ok"
    self.file_selector_cb_button = "ok"
    gtk.mainquit()

  def file_selector_cancel_cb(self, *args):
    #print "cancel"
    self.file_selector_cb_button = "cancel"
    gtk.mainquit()

  def _select_chapter(self, *args):
    if len(self.book.chapters) == 0:
      ButtonChoiceWindow()("Error", "No chapters found!", ["Bummer"],
                         transient=self.win)

    else:
      d = -1
      for c in self.book.chapters:
        if self._offset >= c[1]:
          d = d + 1
    
      button, offset = ListBoxChoiceWindow()("Select Chapter", "",
					     self.book.chapters, default=d)

      if button == "Ok":
        self._chapter_jump(None, offset)

  def _calc_length(self, *args):
    estimate = self.calc_length()
    text =  "Total play time at %s wpm:\n%s hours, %s minutes, %s seconds" % \
           (self.wpm,
            estimate / 3600,
            estimate % 3600 / 60,
            estimate % 60)

    ButtonChoiceWindow()("Calculate Length", text,
                       ["Ok"], transient=self.win)

  def _calc_remaining(self, *args):
    estimate = self.calc_length(self._offset)
    text =  "Remaining play time at %s wpm:\n%s hours, %s minutes, %s seconds" % \
           (self.wpm,
            estimate / 3600,
            estimate % 3600 / 60,
            estimate % 60)

    ButtonChoiceWindow()("Calculate Remaining", text,
                       ["Ok"], transient=self.win)

  def calc_length(self, start=0):
    #print "calc_length(%s)" % start
    #estimate = 60.0 * (len(self.words) - start) / self.wpm

    estimate = 0
    costs = {
      " "        : 1,
      ","       : self.cfg.comma_delay + 2,
      ":"       : self.cfg.colon_delay + 2,
      "."       : self.cfg.sentence_end_delay + 2,
      "\n"      : self.cfg.paragraph_end_delay + 2,
      "c" : self.cfg.chapter_end_delay + 2,
    }
    #for w in self.words:
    for i in xrange(start, self.book.len()):
      t = self.book.type(i)
      estimate = estimate + costs[t]

    estimate = 60.0 * estimate / self.wpm
    estimate = int(estimate)
    return estimate

  def _set_mode(self, data, menuitem):
    modes = ["one word", "three words", "multiword"]
    self.cfg.display_mode = modes[data]

  def _handle_speedbar(self, adj):
    self.wpm = adj.value
    if self.timer:
      self._stop_button()
      self._play_button()

  def _handle_percentbar(self, adj):
    if self._ignore_percent_events:
      return
    percent = adj.value
    #percent = "%.0f%%" % (100.0 * float(self._offset + 1) / len(self.words))
    #self.percent_label.set_text(percent)
    self._offset = int(percent * (self.book.len()-1) / 100.0)
    self.show_word(flush=1)
    
  def _stop_button(self, *args):
    if self.timer:
      gtk.timeout_remove(self.timer)
      self.timer = None
    self._displayed = 0

  def _step_button(self, *args):
    while not self._displayed:
      if self._offset >= self.book.len():
        self._displayed = 1
        break
      self._timer()
    self._stop_button()
    #self._displayed = 0

  def _rplay_button(self, *args):
    # The following sets up reverse play:
    #self.delay_time = int(60.0 / self.wpm * 1000.0)
    #self.direction = -1
    #if not self.timer:
    #  self.timer = gtk.timeout_add(self.delay_time, self._timer, None)

    # But a simple skip-back-by-one-sentence is more useful:
    self._skip_backward([".", "\n", "c"])

  def _play_button(self, *args):
    self.delay_time = int(60.0 / self.wpm * 1000.0)
    self.direction = 1
    if not self.timer:
      self.timer = gtk.timeout_add(self.delay_time, self._timer, None)

  def _skip_button(self, *args):
    "Skip forward to the next paragraph."
    self._skip_forward(["\n", "c"])

  def _rskip_button(self, *args):
    """Skip back to the previous paragraph.
       ... or to the beginning of the current one."""
    self._skip_backward(["\n", "c"])

  def _next_chapter(self, *args):
    "Skip forward to the next chapter."
    self._skip_forward(["c"])

  def _prev_chapter(self, *args):
    "Skip forward to the next chapter."
    self._skip_backward(["c"])

  def _skip_forward(self, look=[".", "\n", "c"]):
    "Skip forward to the next <something>"
    found = 0
    self._offset = self._offset + 1
    while not found:
      if self._offset >= self.book.len():
        self._offset = self.book.len() - 1
        #word = (self.book.word(self._offset), self.book.type(self._offset))
        self.show_word(flush=1)
        return

      #word = (self.book.word(self._offset), self.book.type(self._offset))
      
      if self.book.type(self._offset) in look:
        found = 1
        self._offset = self._offset + 1
        self.show_word(flush=1)
        break

      self._offset = self._offset + 1

    #word = (self.book.word(self._offset), self.book.type(self._offset))
    self.show_word(flush=1)

  def _skip_backward(self, look=[".", "\n", "c"]):
    """Skip back to the previous paragraph.
       ... or to the beginning of the current one."""
    found = 0

    # If in the first few words of the paragraph, skip to previous.
    # Else skip to beginning of current.
    self._offset = self._offset - self.cfg.rskip_tolerance
    while not found:
      if self._offset < 0:
        self._offset = 0
        #word = self.words[self._offset]
        self.show_word(flush=1)
        return

      #word = self.words[self._offset]
      
      if self.book.type(self._offset) in look:
        found = 1
        self._offset = self._offset + 1
        self.show_word(flush=1)
        break

      self._offset = self._offset - 1

    #word = self.words[self._offset]
    self.show_word(flush=1)

  def update_percent(self):    
    self._ignore_percent_events = 1
    #percent = "%.0f%%" % (100.0 * float(self._offset + 1) / self.book.len())
    #self.percent_label.set_text(percent)
    percent = 100.0 * float(self._offset + 1) / self.book.len()
    self.percent_adjustment.set_value(percent)
    self._ignore_percent_events = 0

  def show_word(self, flush=0):

    if self._offset >= self.book.len():
      self._offset = self.book.len() - 1
    w = (self.book.word(self._offset), self.book.type(self._offset))

    self.update_percent()

    ##########  One Word Mode
    if self.cfg.display_mode == "one word":
      if not flush  and  w[1] != " ":  return

      text = "%s" % (self.book.word(self._offset))
      # For fixed-width fonts, try to make sure the total number
      # of characters is an even number.  This keeps letters from
      # jumping between two different positions
      self.set_textarea(text)
    
    ##########  Three Word Mode
    elif self.cfg.display_mode == "three words":
      if w[1] != " ":  return
      
      showing = []
      if self._offset > 0  and  self.book.type(self._offset - 1) == " ":
        showing.append(self.book.word(self._offset - 1))
      else:
        showing.append("")

      if self._offset < self.book.len()  and  w[1] == " ":
        showing.append(self.book.word(self._offset))
      else:
        showing.append("")

      if self._offset < self.book.len() + 1  and  \
         self.book.type(self._offset + 1) == " ":
        showing.append(self.book.word(self._offset + 1))
      else:
        showing.append("")

      text = "%s %s %s" % (showing[0], showing[1], showing[2])
      self.set_textarea(text)

    ##########  Multiword Mode
    elif self.cfg.display_mode == "multiword":

      # Flush is used when jumping through text...
      # It's basically just a hack to make multiword mode display
      # text immediately, rather than buffering until the output is full.
      if flush:
        while self.book.type(self._offset) != " ":
          self._offset = self._offset + 1
          if self._offset >= self.book.len():
            self._offset = self.book.len() - 1
            break
        
        tmp = self._offset
        text, word = self._mw_buildtext()
        self.set_textarea(text)
        self._offset = tmp
        return

      else:
        # Error!  This shouldn't be called without flush!
        return

  def set_textarea(self, text):
    # Ensure fixed-width fonts don't jitter back and forth:
    # all strings must be of even length.
    if len(text) % 2 == 1:
      text = text + " "

    self.textbox.set_text(text)

    # Keep track of whether the screen content actually changed,
    # so we can make sure the step button is doing something.
    self._displayed = 0
    if text != self._last_text:
      self._displayed = 1

    self._last_text = text

  def _mw_buildtext(self):
    if self._offset >= self.book.len():
      self._offset = self.book.len() - 1
    word = (self.book.word(self._offset), self.book.type(self._offset))
    text = ""
    maxlen = self.cfg.multiword_length

    while word[1] == " ":
      text = "%s%s " % (text, word[0])
      if len(text) > maxlen:
        break
      if self._offset + self.direction < self.book.len():
        word2 = (self.book.word(self._offset+1),
                 self.book.type(self._offset+1))
        if len(text) + len(word2[0]) > maxlen:
          break

      self._offset = self._offset + self.direction
      if self._offset < 0:
        self._offset = 0
        self._stop_button()
        break
      elif self._offset >= self.book.len():
        self._offset = self.book.len() - 1
        self._stop_button()
        break
      word = (self.book.word(self._offset), self.book.type(self._offset))

    if text != ""  and  text[-1] == " ": text = text[:-1]

    return (text, word)


  def _mw_timer(self, *args):
    """
    The multiword mode needs a different delay mechanism.
    It's easiest just to give it its own timer callback.
    It uses a post-phrase delay, rather than a post-word delay.
    It's not good when a long phrase gets flashed too fast because
    the next phrase is only one word long.
    
    As a side effect, multiword mode will actually go through text
    slightly faster than the other modes, at the same wpm setting.
    It goes perhaps 5% faster on average.
    """
    text, word = self._mw_buildtext()

    self.update_percent()

    if text != "":
      #print "Display: %s" % (text)
      self.set_textarea(text)

    self._offset = self._offset + self.direction

    if word[1] == " ":
      self._pausecount = 1
    elif word[1] == ",":
      self._pausecount = int(1.5*(self.cfg.comma_delay+1)) + 1
    elif word[1] == ":":  
      self._pausecount = int(1.5*(self.cfg.colon_delay+1)) + 1
    elif word[1] == ".":  
      self._pausecount = int(1.5*(self.cfg.sentence_end_delay+1)) \
                         + 1
    elif word[1] == "\n":  
      self._pausecount = int(1.5*(self.cfg.paragraph_end_delay+1)) \
                         + 1
    elif word[1] == "c":  
      if self.cfg.stop_at_bookmark:
        self._stop_button()
      else:
        self._pausecount = int(1.5*(self.cfg.chapter_end_delay+1)) \
                           + 1
    else:
      print "Unrecognized word style: '%s'" % word[1]
      
    
    return 1

  def _timer(self, *args):
    # Keep index in bounds...
    if self._offset < 0:
      self._offset = 0
      if self.direction < 0:
        self._stop_button()
    elif self._offset >= self.book.len():
      self._offset = self.book.len() - 1
      self._stop_button()
    

    # Do we need to stop briefly?
    if self._pausecount > 0:
      self._pausecount = self._pausecount - 1
      return 1

    if self.cfg.display_mode == "multiword":
      return self._mw_timer()

    word = (self.book.word(self._offset), self.book.type(self._offset))

    if self.cfg.verbose:
      print "Word: %s (%s)" % (word[0], word[1])
    

    # Handle different word types.
    if word[1] == " ":  # text
      self.show_word()

    elif word[1] == ",":
      self.show_word()
      self._pausecount = self.cfg.comma_delay

    elif word[1] == ":":
      self.show_word()
      self._pausecount = self.cfg.colon_delay

    elif word[1] == ".":
      self.show_word()
      self._pausecount = self.cfg.sentence_end_delay

    elif word[1] == "\n":
      self.show_word()
      self._pausecount = self.cfg.paragraph_end_delay

    elif word[1] == "c":
      self.show_word()
      if self.cfg.stop_at_bookmark:
        self._stop_button()
      else:
        self._pausecount = self.cfg.chapter_end_delay

    else:
      print "Unrecognized word style: %s" % (word[1])
      self.show_word()
        
    self._offset = self._offset + self.direction

    return 1

  def make_test_words(self):
    import string

    self.book = plaintext()
    self.book.make_test_words()

def run(args):
  import pycfg
  cfg = pycfg.config(program_name)
  cfg.load()

  testwords = 0
  infiles = []

  for arg in args:
    if arg == "-n"  or  arg == "-e"  or  arg == "--enter-pbreak":
      cfg.enter_is_paragraph_break = 1

    else:
      infiles.append(arg)
    
  if len(args) <= 0:
    testwords = 1

  if infiles == []:
    v = viewer(cfg)
    v.make_test_words()
    v.run()
  else:
    for filename in infiles:
      v = viewer(cfg)
      v.load_plaintext(filename)
      v.run()

if __name__ == "__main__":
  import sys
  run(sys.argv[1:])
