#!/usr/bin/python


#
# RSVP: Rapid Serial Visual Presentation
#
# A text reader designed for reading etexts on linux-based PDAs.
# Displays one word at a time, quickly, in a large font.
# At least, that's the original idea.  More display modes have
# been added for better readability.
#
# Concepts based on GNU sview and Panamedia PanaRead.
#
# Original author: Scott Scriven <gutenflash@xyzz.org>
#
# This program is distributed under the GNU Public License.
#


#
# TODO: add delays for long words.
# TODO: html input (and "Open URL...")
# TODO: configurable punctuation maps (user defines what punctuation
#       counts as "short", "medium", and "long")
# TODO: smallfont text map? (birds eye view)
# TODO: bookmarks?
# TODO: optimize memory use!!!
# -----
# DONE: Chapter chooser window
# DONE: percent meter
# DONE: percent chooser?  (menu worked, but slider is better)
# DONE: automatic chapter detection
#       (implemented by searching for "^CHAPTER" and similar
# DONE: chapter selection
# DONE: 3-label bigtext widget?
#       (used a 3-word single label instead; easier to read!)
#       (multiword mode is even better)
# DONE: Detect and correct cases where the text doesn't fit!
#       (implemented as multiword mode)
# DONE: Make multiword mode's timing dependent on what is visible
#       rather than what is being buffered in the next word set.
#       (painful hairy messy implementation probably required?)
# DONE: display chapters while reading.
#

program_name = "gutenflash"
program_version = "0.8.3"


# GTK setup crap...
# Importing GTK takes too long, especially on an ipaq.
print "Loading GTK..."

from gtk import *

print "Loading GTK...  done."

def runqueue():
  while events_pending():
    mainiteration()

def Run(*args):
  mainloop()

def Quit(*args):
  mainquit()


#
# Config file handling, and defaults...
#

######################################################################
# Config class:
#
def is_config(obj):
  return hasattr(obj,'is_config') and obj.is_config

class config:
  is_config = 1

  def __init__(self):
    import os
    self.options = {}
    self.usrfilename = "%s/.%s.conf" % (os.environ["HOME"], program_name)
    self.globalfilename = "/etc/%s/%s.conf" % (program_name, program_name)
    self.namespace = {
      "options" : self.options,
      }

  ##################################################
  #
  def load(self):
    import os.path
    for name in [self.globalfilename, self.usrfilename]:
      if os.path.exists(name):
        self.parse(name)

  ##################################################
  #
  def parse(self, filename):
    #print "parsing %s..." % filename
    
    try:
      execfile(filename, self.namespace)
    except:
      print "Error in config file %s" % filename
      import traceback, sys
      foo = traceback.format_exception(sys.exc_type,
                                       sys.exc_value,
                                       sys.exc_traceback)
      for item in foo:
        item = item[:-1]
        print item
      print "Press Enter to continue..."
      raw_input()


######################################################################
# Loads regular text files into the required native data structure.
#
class plaintext:

  def __init__(self, filename=None):
    self.filename = filename
    self.text = ""
    self.lines = []
    self.wpl = 64  # words per line
    self._line = -1
    self.words = [] # words in current line
    self.types = []
    self.chapters = []
    self.length = 0

    self.chaptertexts = [
      "CONTENTS",
      "INDEX",
      "FOREWORD",
      "PROLOGUE",
      "CHAPTER ", "Chapter",
      "PART",
      "BOOK",
      "SECTION ", "Section",
      "EPILOGUE",
      "GLOSSARY",
      "APPENDIX",
      ]

    self.abbreviations = [
      "Dr.", "Mr.", "Mrs.", "Ms.",
      #"A.", "B.", "C.", "D.", "E.", "F.", "G.", "H.",       "J.", "K.",
      #"L.", "M.", "N.", "O.", "P.", "Q.", "R.", "S.", "T.", "U.", "V.",
      #"W.", "X.", "Y.", "Z.",
      ]

    self.win = GtkWindow()
    self.vbox = GtkVBox()
    self.win.add(self.vbox)
    self.label = GtkLabel()
    self.progress = GtkProgressBar()
    self.vbox.add(self.label)
    self.vbox.add(self.progress)

  ##################################################
  #
  def load_line(self, line):
    import string
    
    self._line = line
    self.words = string.split(self.lines[self._line], " ")

  ##################################################
  #
  def word(self, num):
    if self.length <= 0:
      return ""
    
    line = num / self.wpl
    offset = num % self.wpl
    
    #print "getword(%s): line(%s/%s)" % (num, line, len(self.lines))
    if line != self._line:
      self.load_line(line)

    #try:
    word = self.words[offset]
    #except:
    #  print "word %s/%s in: %s" % (offset, len(self.words), self.words)
    #  raise "hell"
    
    #print "Word(%s): '%s'" % (num, word)

    return word

  ##################################################
  #
  def type(self, num):
    if self.length <= 0:
      return " "
    return self.types[num]

  ##################################################
  #
  def len(self):
    if self.length <= 0:
      return 1
    else:
      return self.length

  ##################################################
  #
  # TODO: store current line as a list instead of a string,
  #       for easier popping and less regenerating text, and
  #       join it together when it reaches appropriate length...
  #
  def add_word(self, w, t):
    # w must be non-space, non-null
    # t must be 1 letter

    if t == "c":
      self.chapters.append((w, self.length))
      w = "."

    #print "Word(%s): %s (%s)" % (self.length, w, t)

    self.types.append(t)

    if (self.length % self.wpl == 0):  #or (len(self.lines) == 0):
      self.lines.append(w)
    else:
      #self.lines[-1] = "%s %s" % (self.lines[-1], w)
      self.lines[-1] = self.lines[-1] + " " + w

    self.length = self.length + 1

  ##################################################
  #
  def pop_word(self):
    import string
    
    if self.length <= 0:
      return

    #print "pop_word() Length(%s), lines(%s)" % (self.length, len(self.lines))

    if len(self.lines[-1]) == 0:
      #print "Popping line(a): %s" % self.lines[-1]
      self.lines.pop()
      
    elif self.length % self.wpl == 1:
      #print "Popping line(b): %s" % self.lines[-1]
      self.lines.pop()
    #elif self.length % self.wpl == 0:
    #  pass
    else:
      line = self.lines[-1]
      offset = string.rfind(line, " ")
      self.lines[-1] = line[:offset]
      
    self.types.pop()

    self.length = self.length - 1

  ##################################################
  #
  def make_test_words(self):
    words = [
      "Press", ">", "to", "play", "back", "text.", ["."],
      "The", "[]", "button", "stops", "playback.", ["."],
      "Press", ">[]", [","], "to", "step", "one", "frame", "at",
      "a", "time.", ["."],
      "Use", "|<<", [","],"and", ">>|", [","], "to", "skip", "back", "or",
      "forward", "by", "one", "paragraph,", [","], "and", "|<", [","], "to",
      "skip", "back", "by", "one", "sentence.", ["\n"],
      "Use", "the", "top", "slider", "below", "to", "navigate",
      "quickly", "through", "the", "text.", ["."],
      "The", "bottom", "slider", "below", "controls", "the", "playback",
      "speed.", ["\n"],
      ]

    for w in words:
      if type(w) == type("hi"):
        self.add_word(w, " ")
      else:
        self.add_word(".", w[0])

  ##################################################
  #
  def load(self, filename=None, viewer_win=None, enter_is_pbreak=0):
    import string
    import sys

    if filename:
      self.filename = filename

    try:
      if len(self.filename) > 3  and  self.filename[-3:] == ".gz":
        import gzip
        fp = gzip.open(self.filename, "rb")
      else:
        if self.filename == "-":
          fp = sys.stdin
        else:
          fp = open(self.filename, "rb")
    except Exception, msg:
      print "Couldn't load %s: %s" % (self.filename, msg)
      return

    print "Loading %s..." % (self.filename)

    # Set up progress bar.
    self.label.set_text("Loading %s..." % (self.filename))
    self.progress.update(0.0)
    self.progress.set_usize(100, 15)
    self.win.set_transient_for(viewer_win)
    self.win.set_modal(1)
    self.win.show_all()
    runqueue()

    self.types = []
    prev = None

    current_line = 0
    lines = fp.readlines()
    self.text = ""
    numlines = len(lines)
    update_freq = numlines / 200
    if update_freq < 1: update_freq = 1
    #print "freq: %s" % update_freq

    for line in lines:
      # Update the progress bar
      current_line = current_line + 1
      if current_line % update_freq == 0:
        self.progress.update(float(current_line) / numlines)
        runqueue()
      #print "Line: %s" % (line)

      # Check for chapter headings.
      is_chapter = 0
      stripline = string.strip(line)
      for c in self.chaptertexts:
        if stripline[:len(c)] == c:
          is_chapter = 1
      if is_chapter:
        #print "Chapter: %s (%s)" % (line, len(self.wordlist))
        self.add_word(stripline, "c")

        ws = string.split(stripline)
        for w in ws:
          self.add_word(w, " ")
        self.add_word(".", "\n")
        continue

      # Paragraph break signified by blank lines or indentation
      is_paragraph = 0
      if stripline == "":
        # Ignore leading whitespace
        if self.length == 0:
          continue
        # Compensate for texts with blank lines in the middle of
        # a paragraph.  Ignore paragraph breaks in the middle of sentences.
        if prev == " ":
          continue

        is_paragraph = 1

      # paragraph declared by indentation
      if line[:3] == "   "  or  line[0] == "\t":

        # Compensate for indented blocks of text
        if prev != " ":
          is_paragraph = 1
        

      if is_paragraph:
        # Avoid two different pauses in a row
        # Paragraphs and override regular punctuation
        if prev in [".", ",", "-", ":"]: # prev != ""
          self.pop_word()

        # Chapters override paragraphs.
        # Avoid repeating paragraph breaks.
        if prev not in ["c", "\n"]:
          self.add_word(".", "\n")
          prev = "\n"

      # Parse the words...
      words = string.split(line)
      for w in words:
        #w = string.strip(w)
        #if w == "": continue

        # Re-join hyphen-ated wo-rds
        # FIXME: Broken right now...
        if prev == "-":
          #w2 = self.word(self.length - 1)
          #w2 = w2[:-1]
          #self.pop_word()
          #w = "%s%s" % (w2, w)
          self.add_word(w, " ")
        else: # Regular word
          # FIXME: make double-hyphenation check faster!
          # Check for internal double-hyphenation  "better--and"
          # ("-" check is to eliminate need for "--" check in most words)
          #if "-" in w  and  (string.find(w, "--") > 0):
          #  i1 = string.find(w, "--")
          #  i2 = string.rfind(w, "--") + 2
          #  w1 = w[:i1]
          #  w2 = w[i1:i2]
          #  w3 = w[i2:]
          #  self.add_word(w1, " ")
          #  self.add_word(w2, " ")
          #  self.add_word(w3, " ")
          #  w = w3
	  #  if w == "":  # "sometimes I think--"
	  #    continue
          # Otherwise, regular word
          #else:
          #  self.add_word(w, " ")
          self.add_word(w, " ")
          
        prev = " "

        # Handle punctuation.
        if w[-1] in '.!?"':
          if w not in self.abbreviations:
            self.add_word(".", ".")
            prev = "."
        elif w[-1] in ",;":
          self.add_word(".", ",")
          prev = ","
        elif w[-1] in ":\'\")]}>":
          self.add_word(".", ":")
          prev = ":"
        # Setup for de-hyphenation of words.
        elif w[-1] == "-":
          if len(w) > 1  and  w[-2] in "-":  # "sometimes I think--"
            self.add_word(".", ":")
            prev = ":"
          elif len(w) > 1: #  "then he per- \n formed the operation"
            prev = "-"
          else: # just a dash
            pass

      # Handle paragraph breaks denoted by a \n
      if enter_is_pbreak:
        # Avoid two different pauses in a row
        # Paragraphs and override regular punctuation
        if prev in [".", ",", "-", ":"]: # prev != ""
          self.pop_word()

        # Chapters override paragraphs.
        # Avoid repeating paragraph breaks.
        if prev not in ["c", "\n"]:
          self.add_word(".", "\n")
          prev = "\n"

    self.types = string.join(self.types, "")

    fp.close()
    self.win.hide()
    self.win.unrealize()
    runqueue()
    
    print "Loading %s...  Done." % (self.filename)

  ##################################################
  #
  def __del__(self, *args):
    #self.win.hide()
    #runqueue()
    pass

######################################################################
# Not used.
# Instead, a simple 2-tuple is used.
# It uses less than half as much memory!
# (and this is intended for use on an ipaq, with limited resources)
#
#class word:
#
#  def __init__(self, value, style=""):
#    self.value = value
#    # Possible styles:
#    # - ,
#    # - :
#    # - .
#    # - chapter
#    # - \n
#    self.style = style
#    if not value:
#      self.value = ""



def set_gdk_win_prop(window, property, type,
                     format, mode, data):
  import _gtk, GDK
  atom = _gtk.gdk_atom_intern(property, 1)
  if not int(atom): return 0
  int_atom = _gtk.gdk_atom_intern(type, 1)
  if not int(int_atom): return 0
  if mode=="REPLACE":gdk_mode = GDK.PROP_MODE_REPLACE
  elif mode=="APPEND":gdk_mode = GDK.PROP_MODE_APPEND
  elif mode=="PREPEND":gdk_mode = GDK.PROP_MODE_PREPEND
  try:
    window.property_change(atom, int_atom, format, gdk_mode, data)
    return 1
  except: return 0



############################################################
#
#
_ButtonChoice_clicked = None

def _ButtonChoiceButtonCB(button):
  global _ButtonChoice_clicked
  _ButtonChoice_clicked = button
  mainquit()


############################################################
#
#
def ButtonChoiceWindow(title, text, buttons, default=0, long=0,
                       transient=None):
  global _ButtonChoice_clicked
  _ButtonChoice_clicked = None

  win = GtkWindow()
  if transient:
    win.set_transient_for(transient)
  win.connect("destroy", mainquit)
  win.set_position(WIN_POS_CENTER)
  win.set_modal(1)
  win.set_title(title)
  win.set_border_width(5)

  v = GtkVBox()

  #if len(text) < 300:
  if not long:

    l = GtkLabel(text)
    l.set_line_wrap(1)
    l.show()
    v.add(l)
    
  else:  #if len(text) > 300:
    
    lbox = GtkHBox()
    l = GtkText(None, None)
    l.set_word_wrap(1)
    l.set_line_wrap(1)
    scroll = GtkVScrollbar(l.get_vadjustment())
    lbox.add(l)
    lbox.add(scroll)
    l.insert(None, None, None, text)
    l.set_usize(400, 300)
    v.add(lbox)
    

  box = GtkHButtonBox()
  box.set_border_width(10)


  buttonlist = []
  i = 0
  for item in buttons:
    if type(item) == type((1,2)):
      label, value = item
    else:
      label = value = item


    b = GtkButton(label)
    buttonlist.append([label, value, b])
    b.connect("clicked", _ButtonChoiceButtonCB)

    if default >= 0  and  i == default:
      b.set_flags(CAN_DEFAULT)
      b.set_flags(HAS_DEFAULT)
      box.add(b)
    #if i != default  and  default >= 0:
    #  #box.add(b)
    #  f = GtkFrame()
    #  f.set_border_width(2)
    #  f.set_shadow_type(SHADOW_NONE)
    #  f.add(b)
    #  box.add(f)
    else:
      f = GtkFrame()
      f.set_border_width(2)
      f.set_shadow_type(SHADOW_NONE)
      f.add(b)
      box.add(f)

    i = i + 1

  box.show()
  v.add(box)

  v.show()
  win.add(v)

  win.show_all()
  mainloop()

  retval = None

  for item in buttonlist:
    label, value, b = item
    if b == _ButtonChoice_clicked:
      retval = value

  win.hide()
  win.unrealize()
  del win

  return retval



############################################################
#
#
def ListBoxChoiceWindow(title, text, items, buttons = ["Ok", "Cancel"],
                        default=0):
  global _ListBoxChoiceWindow_clicked
  _ListBoxChoiceWindow_clicked = None

  win = GtkWindow()
  win.set_position(WIN_POS_CENTER)
  win.connect("destroy", mainquit)
  win.set_title(title)
  win.set_modal(1)

  v = GtkVBox()
  win.add(v)

  l = GtkLabel(text)
  l.set_line_wrap(1)
  #l.set_padding(0)
  v.add(l, fill=0, expand=0)

  box = ListBox()
  box.set_padding(0)
  box.add(items)
  box.default(default)
  v.add(box.w, fill=1, expand=1)

  bbar = ButtonBar(buttons)
  v.add(bbar.w, fill=0, expand=0)
  

  win.show_all()
  mainloop()
  win.hide()

  retval = None

  selected = box.get_value()
  clicked = bbar.get_value()

  retval = (clicked, selected)

  win.unrealize()

  return retval



############################################################
# Widget class...
#
# Conventions:
#  - classes derived from Widget should name their primary GTK+
#    widget "self.w".  This allows callbacks to work.
#  - classes derived from Widget should call "self._init()" in
#    their __init__().
#
class Widget:

  def __init__(self):
    self._init()

  def _init(self):
    self._cb = {}
    self._cb_func = {
      "pressed" : self._pressed_cb,
      "released" : self._released_cb,
      "clicked" : self._clicked_cb,
      "toggled" : self._toggled_cb,
      "enter" : self._enter_cb,
      "leave" : self._leave_cb,
      "destroy" : self._destroy_cb,
      "" : None
      }


  def _pressed_cb(self, *args):
    self._callback("pressed")

  def _released_cb(self, *args):
    self._callback("released")

  def _clicked_cb(self, *args):
    self._callback("clicked")

  def _toggled_cb(self, *args):
    global _object_table
    (widget, ) = args
    value = widget.get_active()
    self._callback("toggled", value)

  def _enter_cb(self, *args):
    self._callback("enter")

  def _leave_cb(self, *args):
    self._callback("leave")

  def _destroy_cb(self, *args):
    self._callback("destroy")

  def _callback(self, event, *args):
    if self._cb.has_key(event):
      parameters = (self, event) + args
      #print "Calling %s cb...  Args: %s" % (event, repr(parameters))
      apply(self._cb[event], parameters)
      #print "Calling %s cb... done.\n" % event
  
  def set_cb(self, event, cb):
    if self._cb_func.has_key(event)  and  hasattr(self, "w"):
      self._cb[event] = cb
      self.w.connect(event, self._cb_func[event])
    else:
      raise ValueError, "Invalid event type %s" % event

  def run(self):
    mainloop()

  def show(self):
    if hasattr(self, "w"):
      self.w.show()
    
  def hide(self):
    if hasattr(self, "w"):
      self.w.hide()

  def set_padding(self, xpad, ypad=None):
    if ypad == None:
      ypad = xpad
    if hasattr(self, "w"):
      if hasattr(self.w, "set_padding"):
        self.w.set_padding(xpad, ypad)
      elif hasattr(self.w, "set_border_width"):
        self.w.set_border_width(xpad)
    

######################################################################
#
#
class Button(Widget):

  def __init__(self, text=None):
    self._init()
    self.w = GtkButton()
    if text:
      self.set_text(text)

  def set_text(self, text):
    if hasattr(self, "l"):
      self.l.set_text(text)
    else:
      self.l = GtkLabel()
      self.l.set_line_wrap(1)
      self.l.set_padding(5, 3)
      self.l.set_text(text)
      self.w.add(self.l)


######################################################################
#
#
class ButtonBar(Widget):

  def __init__(self, buttons=[]):
    self._init()
    self.w = GtkHButtonBox()
    self.w.set_layout_default(BUTTONBOX_SPREAD)
    self.buttons = []
    self._clicked = None
    for button in buttons:
      if type(button) == type((1,2)):
        text, value = button
      else:
        text = value = button
      b = Button(text)
      b.set_cb("clicked", self._default_cb)
      self.buttons.append((b, value))
      self.w.add(b.w)

  def set_cb(self, which, cb):
    try:
      (button, value) = self.buttons[which]
    except:
      raise IndexError, "Button out of range"
    button.set_cb("clicked", cb)

  def _default_cb(self, object, event):
    for item in self.buttons:
      (button, value) = item
      if button == object:
        self._clicked = value
        Quit()

  def get_value(self, result=None):
    return self._clicked


######################################################################
#
#
class ListBox(Widget):

  def __init__(self, height=10):
    self._init()
    self.w = GtkScrolledWindow()
    self.w.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)
    self.w.set_border_width(2)
    self.l = GtkCList(1)
    self.l.columns_autosize()
    #self.l.set_usize(200, 20 + (height * 12))
    self.items = []
    self.l.connect("select-row", self._item_select_cb)
    self.l.set_selection_mode(SELECTION_BROWSE)
    self.w.add(self.l)
    
  def add_item(self, item):
    if type(item) == type((1,2)):
      text, value = item
    else:
      text = value = item
    
    self.items.append((text, value, 0))
    self.l.insert(len(self.items), [text])

  def add(self, items):
    if type(items) == type([1, 2]):
      for item in items:
        self.add_item(item)
    else:
      self.add_item(items)

  def get_value(self):
    result = []
    for item in self.items:
      (text, value, on) = item
      if on:
        result.append(value)

    if len(result) > 0:
      return result[0]
    else:
      return None
    

  def _item_select_cb(self, clist, row, col, gdkevent):
    self.deselect_all()
    (text, value, on) = self.items[row]
    self.items[row] = (text, value, 1)

  def default(self, which):
    self.deselect_all()
    self.select(which)
    self.l.moveto(which, 1, 0.5, 0.5)

  def deselect_all(self):
    for i in range(len(self.items)):
      (text, value, on) = self.items[i]
      self.items[i] = (text, value, 0)

  # TODO: Make sure selected item is actually visible  (scroll to it)
  def select(self, which):
    if which >= 0  and  which < len(self.items):
      (text, value, on) = self.items[which]
      self.l.select_row(which, 0)
      self.items[which] = (text, value, 1)

      
######################################################################
#
class viewer:

  def __init__(self, cfg):
    self.cfg = cfg
    self.book = None
    self.win = GtkWindow()
    self.win.connect("destroy", self._destroy)
    self.vbox = GtkVBox()
    self.win.add(self.vbox)

    self.make_menus()

    self.make_textbox()
    
    self.wpm = self.cfg.options["words per minute"]

    self.direction = 1

    self.make_speed_bar()
    self.make_play_buttons()

    self.filename = ""
    self.shortfilename = ""

    self._offset = 0
    self._pausecount = 0

    self._docked = 0
    if self.cfg.options["toolbar app"]:
      self.set_docked(1, show=0)
    self.win.set_title("%s %s" % (program_name, program_version))
    self.win.show_all()
    runqueue()


  ##################################################
  #
  def nop(self, *args):
    pass

  ##################################################
  #
  def _toggle_docked(self, *args):
    self.set_docked(not self._docked)

  ##################################################
  #
  def set_docked(self, dock, show=1):
    import _gtk
    
    self.win.realize()
    if show:
      self.win.hide()
    if dock:
      t = [_gtk.gdk_atom_intern("_NET_WM_WINDOW_TYPE_TOOLBAR")]
    else:
      t = []
    set_gdk_win_prop(self.win.get_window(), "_NET_WM_WINDOW_TYPE",
                     "ATOM", 32, "REPLACE", t)
    if show:
      # FIXME: can't resize when docking
      #self.textbox.set_usize(self.cfg.options["view wid"],
      #                       self.cfg.options["view hgt"])
      self.win.show_all()
    self._docked = dock
    self._displayed = 0

  ##################################################
  #
  def load_plaintext(self, infile):
    import string
    self.filename = infile
    index = string.rfind(infile, "/") + 1
    self.shortfilename = infile[index:]
    
    self.book = plaintext()
    self.book.load(infile, self.win,
                   self.cfg.options["enter is paragraph break"])
    #self.update_chapter_menu()
    self.win.set_title("%s: %s" % (program_name, self.shortfilename))
    #import sys
    #sys.exit(0)

  ##################################################
  #
  def make_menus(self):
    self.ag = GtkAccelGroup()
    self.win.add_accel_group(self.ag)
    self.itemf = GtkItemFactory(GtkMenuBar, "<main>", self.ag)
    self.itemf.create_items([
      ("/_File",       None, None, 0, "<Branch>"),
      ("/File/_Open",  "<Control>O", self._file_open, 0, ""),
      ("/File/Calculate _Length",  "<Control>L", self._calc_length, 0, ""),
      ("/File/Calculate _Remaining",  "<Control>R", self._calc_remaining, 0, ""),
      #("/File/Toggle Dock" ,  None, self._toggle_docked, 0, ""),
      ("/File/_Quit",  "Q", self._destroy, 0, ""),
      ("/_Bookmarks",   None, None, 0, "<Branch>"),
      #("/_Chapters",   None, None, 0, "<Branch>"),
      ("/_Jump",       None, None, 0, "<Branch>"),
      ("/Jump/To _Chapter...",  "C", self._select_chapter, 0, ""),
      ("/Jump/Next Chapter",  None, self._next_chapter, 0, ""),
      ("/Jump/Previous Chapter",  None, self._prev_chapter, 0, ""),
      ("/Jump/Start",  None, self._percent_jump, 0, ""),
      ("/Jump/End",    None, self._percent_jump, 100, ""),
      ("/_Mode",       None, None, 0, "<Branch>"),
      ("/Mode/One Word",    None, self._set_mode, 0, ""),
      ("/Mode/Three Words", None, self._set_mode, 1, ""),
      ("/Mode/Multiword",   None, self._set_mode, 2, ""),
      ("/_About",       None, None, 0, "<Branch>"),
      ("/About/About %s" % (program_name), None, self.about_screen, 0, ""),
      ])
    self.menubar = self.itemf.get_widget("<main>")
    #self.chapter_menu = self.itemf.get_widget("/Chapters")
    self.vbox.pack_start(self.menubar, fill=FALSE, expand=FALSE)

    foo = GtkItemFactory(GtkMenuBar, "<keydefs>", self.ag)
    foo.create_items([
      ("/foo",       None, None, 0, "<Branch>"),
      ("/foo/rskip", "B", self._rskip_button, 0, ""),
      ("/foo/rplay", "R", self._rplay_button, 0, ""),
      ("/foo/stop",  "S", self._stop_button, 0, ""),
      ("/foo/step",  "E", self._step_button, 0, ""),
      ("/foo/play",  "P", self._play_button, 0, ""),
      ("/foo/skip",  "F", self._skip_button, 0, ""),
      ])
    

  ##################################################
  #
  def update_chapter_menu(self):

    for c in self.book.chapters:
      #print "Chapter(%s): %s" % (c[1], c[0])
      mi = GtkMenuItem(c[0])
      mi.connect("activate", self._chapter_jump, c[1])
      self.chapter_menu.append(mi)

  ##################################################
  #
  def make_textbox(self):
    self._last_text = ""
    self.textbox = GtkLabel()
    self.textbox.set_usize(self.cfg.options["view wid"],
                         self.cfg.options["view hgt"])
    self.font = load_font(self.cfg.options["bigfont"])
    style = self.textbox.get_style()
    style.font = self.font
    self.textbox.set_style(style)
    self.timer = None

    self.vbox.pack_start(self.textbox, fill=TRUE, expand=TRUE)

    

  ##################################################
  #
  def make_play_buttons(self):
    self.button_box = GtkHBox()
    self.rskip_button = GtkButton("|<<")
    self.button_box.add(self.rskip_button)
    self.rplay_button = GtkButton("|<")
    self.button_box.add(self.rplay_button)
    self.stop_button = GtkButton("[]")
    self.button_box.add(self.stop_button)
    self.step_button = GtkButton(">[]")
    self.button_box.add(self.step_button)
    self.play_button = GtkButton(">")
    self.button_box.add(self.play_button)
    self.skip_button = GtkButton(">>|")
    self.button_box.add(self.skip_button)
    self.vbox.pack_start(self.button_box, fill=TRUE, expand=FALSE)

    self.stop_button.connect("clicked", self._stop_button)
    self.step_button.connect("clicked", self._step_button)
    self.play_button.connect("clicked", self._play_button)
    self.rplay_button.connect("clicked", self._rplay_button)
    self.skip_button.connect("clicked", self._skip_button)
    self.rskip_button.connect("clicked", self._rskip_button)

  ##################################################
  #
  def make_speed_bar(self):
    self._ignore_percent_events = 0
    self.percent_adjustment = GtkAdjustment(0.0, 0.0, 100.0,
                                          0.1, 5.0, 0.0)
    self.percent_adjustment.set_value(0.0)
    self.percent_adjustment.connect("value-changed", self._handle_percentbar)
    self.percentbar = GtkHScale(self.percent_adjustment)
    self.percentbar.set_draw_value(1)
    self.percentbar.set_digits(1)
    self.vbox.add(self.percentbar, fill=TRUE, expand=FALSE)
    
    #self.speed_hbox = GtkHBox()
    #self.vbox.pack_start(self.speed_hbox, fill=TRUE, expand=FALSE)
    self.speed_adjustment = GtkAdjustment(self.wpm, 20.0, 2000.0,
                                          5.0, 50.0, 0.0)
    self.speed_adjustment.set_value(self.wpm)
    self.speed_adjustment.connect("value-changed", self._handle_speedbar)
    self.speedbar = GtkHScale(self.speed_adjustment)
    self.speedbar.set_draw_value(1)
    self.speedbar.set_digits(0)
    #self.speedbar.set_value_pos(POS_RIGHT)
    self.vbox.add(self.speedbar, fill=TRUE, expand=FALSE)

    #self.vbox.pack_start(self.speedbar, fill=TRUE, expand=FALSE)
    #self.percent_label = GtkLabel("0%")
    #self.speed_hbox.pack_start(self.speedbar, fill=TRUE, expand=TRUE)
    #self.speed_hbox.pack_start(self.percent_label, fill=FALSE, expand=FALSE)


  ##################################################
  #
  def run(self):
    self.win.show_all()
    self.show_word(flush=1)
    #self.win.set_title("%s: %s" % (program_name, self.shortfilename))
    mainloop()

  ##################################################
  #
  def about_screen(self, *args):
    self.about_win = GtkWindow()
    self.about_win.connect("destroy", self._about_destroy)
    vbox = GtkVBox()
    self.about_win.add(vbox)

    text = \
"""%s V%s

This is a text reader based on the concept of RSVP.  Rapid Serial Visual Presentation.  The idea is to display one word at a time, very quickly and in a large font.

At least, that's the original idea.  More display modes have been added for better readability.

Concepts based on GNU sview and Panamedia PanaRead.

Original author: Scott Scriven <gutenflash@xyzz.org>

This program is distributed under the GNU Public License.
""" % (program_name, program_version)
    hbox = GtkHBox()
    vbox.add(hbox, fill=1, expand=1)
    textbox = GtkText()
    textbox.set_word_wrap(1)
    vscroll = GtkVScrollbar(textbox.get_vadjustment())
    hbox.add(textbox, fill=1, expand=1)
    hbox.add(vscroll, fill=0, expand=0)
    textbox.insert_defaults(text)
    
    closebut = GtkButton("Close")
    closebut.connect("clicked", self._about_destroy)
    vbox.add(closebut, fill=1, expand=0)

    self.about_win.show_all()
    mainloop()

  def _about_destroy(self, *args):
    mainquit()
    self.about_win.hide()
    self.about_win.unrealize()
    del self.about_win

  ##################################################
  #
  def _destroy(self, *args):
    self.win.hide()
    self.win.unrealize()
    del self.win
    mainquit()
    for key in self.__dict__.keys():
      del self.__dict__[key]

  ##################################################
  #
  def _chapter_jump(self, menuitem, data):
    self._offset = data
    self.show_word(flush=1)
    self._displayed = 0

  ##################################################
  #
  def _percent_jump(self, data, menuitem):
    self._offset = int(data * self.book.length / 100.0)
    self.show_word(flush=1)

  ##################################################
  #
  def _file_open(self, *args):
    filename = self.file_selector("Open...", self.cfg.options["load dir"])
    if filename:
      v = viewer(self.cfg)
      self._destroy()
      v.load_plaintext(filename)
      v.run()
    

  ##################################################
  #
  def file_selector(self, title, filename):
    self.file_selector_cb_button = None
    
    fs = GtkFileSelection(title)
    if filename:
      fs.set_filename(filename)
    fs.ok_button.connect("clicked", self.file_selector_ok_cb)
    fs.cancel_button.connect("clicked", self.file_selector_cancel_cb)
    fs.show_all()
    mainloop()

    if self.file_selector_cb_button == "ok":
      filename = fs.get_filename()
      fs.hide()
      fs.unrealize()
      return filename
    else:
      fs.hide()
      fs.unrealize()
      return None

  def file_selector_ok_cb(self, *args):
    #print "ok"
    self.file_selector_cb_button = "ok"
    mainquit()

  def file_selector_cancel_cb(self, *args):
    #print "cancel"
    self.file_selector_cb_button = "cancel"
    mainquit()

  ##################################################
  #
  def _select_chapter(self, *args):
    if len(self.book.chapters) == 0:
      ButtonChoiceWindow("Error", "No chapters found!", ["Bummer"],
                         transient=self.win)

    else:
      d = -1
      for c in self.book.chapters:
        if self._offset >= c[1]:
          d = d + 1
    
      button, offset = ListBoxChoiceWindow("Select Chapter", "",
                                           self.book.chapters, default=d)

      if button == "Ok":
        self._chapter_jump(None, offset)

  ##################################################
  #
  def _calc_length(self, *args):
    estimate = self.calc_length()
    text =  "Total play time at %s wpm:\n%s hours, %s minutes, %s seconds" % \
           (self.wpm,
            estimate / 3600,
            estimate % 3600 / 60,
            estimate % 60)

    ButtonChoiceWindow("Calculate Length", text,
                       ["Ok"], transient=self.win)

  ##################################################
  #
  def _calc_remaining(self, *args):
    estimate = self.calc_length(self._offset)
    text =  "Remaining play time at %s wpm:\n%s hours, %s minutes, %s seconds" % \
           (self.wpm,
            estimate / 3600,
            estimate % 3600 / 60,
            estimate % 60)

    ButtonChoiceWindow("Calculate Remaining", text,
                       ["Ok"], transient=self.win)

  ##################################################
  #
  def calc_length(self, start=0):
    #print "calc_length(%s)" % start
    #estimate = 60.0 * (len(self.words) - start) / self.wpm

    estimate = 0
    costs = {
      " "        : 1,
      ","       : self.cfg.options["comma delay"] + 2,
      ":"       : self.cfg.options["colon delay"] + 2,
      "."       : self.cfg.options["sentence end delay"] + 2,
      "\n"      : self.cfg.options["paragraph end delay"] + 2,
      "c" : self.cfg.options["chapter end delay"] + 2,
    }
    #for w in self.words:
    for i in xrange(start, self.book.len()):
      t = self.book.type(i)
      estimate = estimate + costs[t]

    estimate = 60.0 * estimate / self.wpm
    estimate = int(estimate)
    return estimate

  ##################################################
  #
  def _set_mode(self, data, menuitem):
    modes = ["one word", "three words", "multiword"]
    self.cfg.options["display mode"] = modes[data]

  ##################################################
  #
  def _handle_speedbar(self, adj):
    self.wpm = adj.value
    if self.timer:
      self._stop_button()
      self._play_button()

  ##################################################
  #
  def _handle_percentbar(self, adj):
    if self._ignore_percent_events:
      return
    percent = adj.value
    #percent = "%.0f%%" % (100.0 * float(self._offset + 1) / len(self.words))
    #self.percent_label.set_text(percent)
    self._offset = int(percent * (self.book.len()-1) / 100.0)
    self.show_word(flush=1)
    

  ##################################################
  #
  def _stop_button(self, *args):
    if self.timer:
      timeout_remove(self.timer)
      self.timer = None
    self._displayed = 0

  ##################################################
  #
  def _step_button(self, *args):
    while not self._displayed:
      if self._offset >= self.book.len():
        self._displayed = 1
        break
      self._timer()
    self._stop_button()
    #self._displayed = 0

  ##################################################
  #
  def _rplay_button(self, *args):
    # The following sets up reverse play:
    #self.delay_time = 60.0 / self.wpm * 1000.0
    #self.direction = -1
    #if not self.timer:
    #  self.timer = timeout_add(self.delay_time, self._timer, None)

    # But a simple skip-back-by-one-sentence is more useful:
    self._skip_backward([".", "\n", "c"])

  ##################################################
  #
  def _play_button(self, *args):
    self.delay_time = 60.0 / self.wpm * 1000.0
    self.direction = 1
    if not self.timer:
      self.timer = timeout_add(self.delay_time, self._timer, None)

  ##################################################
  # Skip forward to the next paragraph.
  #
  def _skip_button(self, *args):
    self._skip_forward(["\n", "c"])

  ##################################################
  # Skip back to the previous paragraph.
  # ... or to the beginning of the current one.
  #
  def _rskip_button(self, *args):
    self._skip_backward(["\n", "c"])

  ##################################################
  # Skip forward to the next chapter.
  #
  def _next_chapter(self, *args):
    self._skip_forward(["c"])

  ##################################################
  # Skip forward to the next chapter.
  #
  def _prev_chapter(self, *args):
    self._skip_backward(["c"])

  ##################################################
  # Skip forward to the next <something>
  #
  def _skip_forward(self, look=[".", "\n", "c"]):
    found = 0
    self._offset = self._offset + 1
    while not found:
      if self._offset >= self.book.len():
        self._offset = self.book.len() - 1
        #word = (self.book.word(self._offset), self.book.type(self._offset))
        self.show_word(flush=1)
        return

      #word = (self.book.word(self._offset), self.book.type(self._offset))
      
      if self.book.type(self._offset) in look:
        found = 1
        self._offset = self._offset + 1
        self.show_word(flush=1)
        break

      self._offset = self._offset + 1

    #word = (self.book.word(self._offset), self.book.type(self._offset))
    self.show_word(flush=1)

  ##################################################
  # Skip back to the previous paragraph.
  # ... or to the beginning of the current one.
  #
  def _skip_backward(self, look=[".", "\n", "c"]):
    found = 0

    # If in the first few words of the paragraph, skip to previous.
    # Else skip to beginning of current.
    self._offset = self._offset - self.cfg.options["rskip tolerance"]
    while not found:
      if self._offset < 0:
        self._offset = 0
        #word = self.words[self._offset]
        self.show_word(flush=1)
        return

      #word = self.words[self._offset]
      
      if self.book.type(self._offset) in look:
        found = 1
        self._offset = self._offset + 1
        self.show_word(flush=1)
        break

      self._offset = self._offset - 1

    #word = self.words[self._offset]
    self.show_word(flush=1)

  ##################################################
  #
  def update_percent(self):    
    self._ignore_percent_events = 1
    #percent = "%.0f%%" % (100.0 * float(self._offset + 1) / self.book.len())
    #self.percent_label.set_text(percent)
    percent = 100.0 * float(self._offset + 1) / self.book.len()
    self.percent_adjustment.set_value(percent)
    self._ignore_percent_events = 0

  ##################################################
  #
  def show_word(self, flush=0):

    if self._offset >= self.book.len():
      self._offset = self.book.len() - 1
    w = (self.book.word(self._offset), self.book.type(self._offset))

    self.update_percent()

    ##########  One Word Mode
    if self.cfg.options["display mode"] == "one word":
      if not flush  and  w[1] != " ":  return

      text = "%s" % (self.book.word(self._offset))
      # For fixed-width fonts, try to make sure the total number
      # of characters is an even number.  This keeps letters from
      # jumping between two different positions
      self.set_textarea(text)
    
    ##########  Three Word Mode
    elif self.cfg.options["display mode"] == "three words":
      if w[1] != " ":  return
      
      showing = []
      if self._offset > 0  and  self.book.type(self._offset - 1) == " ":
        showing.append(self.book.word(self._offset - 1))
      else:
        showing.append("")

      if self._offset < self.book.len()  and  w[1] == " ":
        showing.append(self.book.word(self._offset))
      else:
        showing.append("")

      if self._offset < self.book.len() + 1  and  \
         self.book.type(self._offset + 1) == " ":
        showing.append(self.book.word(self._offset + 1))
      else:
        showing.append("")

      text = "%s %s %s" % (showing[0], showing[1], showing[2])
      self.set_textarea(text)

    ##########  Multiword Mode
    elif self.cfg.options["display mode"] == "multiword":

      # Flush is used when jumping through text...
      # It's basically just a hack to make multiword mode display
      # text immediately, rather than buffering until the output is full.
      if flush:
        while self.book.type(self._offset) != " ":
          self._offset = self._offset + 1
          if self._offset >= self.book.len():
            self._offset = self.book.len() - 1
            break
        
        tmp = self._offset
        text, word = self._mw_buildtext()
        self.set_textarea(text)
        self._offset = tmp
        return

      else:
        # Error!  This shouldn't be called without flush!
        return

  ##################################################
  #
  def set_textarea(self, text):
    # Ensure fixed-width fonts don't jitter back and forth:
    # all strings must be of even length.
    if len(text) % 2 == 1:
      text = text + " "

    self.textbox.set_text(text)

    # Keep track of whether the screen content actually changed,
    # so we can make sure the step button is doing something.
    self._displayed = 0
    if text != self._last_text:
      self._displayed = 1

    self._last_text = text


  ##################################################
  #
  def _mw_buildtext(self):
    if self._offset >= self.book.len():
      self._offset = self.book.len() - 1
    word = (self.book.word(self._offset), self.book.type(self._offset))
    text = ""
    maxlen = self.cfg.options["multiword length"]

    while word[1] == " ":
      text = "%s%s " % (text, word[0])
      if len(text) > maxlen:
        break
      if self._offset + self.direction < self.book.len():
        word2 = (self.book.word(self._offset+1),
                 self.book.type(self._offset+1))
        if len(text) + len(word2[0]) > maxlen:
          break

      self._offset = self._offset + self.direction
      if self._offset < 0:
        self._offset = 0
        self._stop_button()
        break
      elif self._offset >= self.book.len():
        self._offset = self.book.len() - 1
        self._stop_button()
        break
      word = (self.book.word(self._offset), self.book.type(self._offset))

    if text != ""  and  text[-1] == " ": text = text[:-1]

    return (text, word)


  ##################################################
  # The multiword mode needs a different delay mechanism.
  # It's easiest just to give it its own timer callback.
  # It uses a post-phrase delay, rather than a post-word delay.
  # It's not good when a long phrase gets flashed too fast because
  # the next phrase is only one word long.
  #
  # As a side effect, multiword mode will actually go through text
  # slightly faster than the other modes, at the same wpm setting.
  # It goes perhaps 5% faster on average.
  #
  def _mw_timer(self, *args):
    text, word = self._mw_buildtext()

    self.update_percent()

    if text != "":
      #print "Display: %s" % (text)
      self.set_textarea(text)

    self._offset = self._offset + self.direction

    if word[1] == " ":
      self._pausecount = 1
    elif word[1] == ",":
      self._pausecount = int(1.5*(self.cfg.options["comma delay"]+1)) + 1
    elif word[1] == ":":  
      self._pausecount = int(1.5*(self.cfg.options["colon delay"]+1)) + 1
    elif word[1] == ".":  
      self._pausecount = int(1.5*(self.cfg.options["sentence end delay"]+1)) \
                         + 1
    elif word[1] == "\n":  
      self._pausecount = int(1.5*(self.cfg.options["paragraph end delay"]+1)) \
                         + 1
    elif word[1] == "c":  
      if self.cfg.options["stop at bookmark"]:
        self._stop_button()
      else:
        self._pausecount = int(1.5*(self.cfg.options["chapter end delay"]+1)) \
                           + 1
    else:
      print "Unrecognized word style: '%s'" % word[1]
      
    
    return 1

  ##################################################
  #
  def _timer(self, *args):

    # Keep index in bounds...
    if self._offset < 0:
      self._offset = 0
      if self.direction < 0:
        self._stop_button()
    elif self._offset >= self.book.len():
      self._offset = self.book.len() - 1
      self._stop_button()
    

    # Do we need to stop briefly?
    if self._pausecount > 0:
      self._pausecount = self._pausecount - 1
      return 1

    if self.cfg.options["display mode"] == "multiword":
      return self._mw_timer()

    word = (self.book.word(self._offset), self.book.type(self._offset))

    if self.cfg.options["verbose"]:
      print "Word: %s (%s)" % (word[0], word[1])
    

    # Handle different word types.
    if word[1] == " ":  # text
      self.show_word()

    elif word[1] == ",":
      self.show_word()
      self._pausecount = self.cfg.options["comma delay"]

    elif word[1] == ":":
      self.show_word()
      self._pausecount = self.cfg.options["colon delay"]

    elif word[1] == ".":
      self.show_word()
      self._pausecount = self.cfg.options["sentence end delay"]

    elif word[1] == "\n":
      self.show_word()
      self._pausecount = self.cfg.options["paragraph end delay"]

    elif word[1] == "c":
      self.show_word()
      if self.cfg.options["stop at bookmark"]:
        self._stop_button()
      else:
        self._pausecount = self.cfg.options["chapter end delay"]

    else:
      print "Unrecognized word style: %s" % (word[1])
      self.show_word()
        
    self._offset = self._offset + self.direction

    return 1

  ##################################################
  #
  def make_test_words(self):
    import string

    self.book = plaintext()
    self.book.make_test_words()
#    self.book.text = """Orbiting this at a distance of roughly ninety-two
#million miles is an utterly insignificant little blue planet whose
#ape-descended life forms are so amazingly primitive that they still think
#digital watches are a pretty neat idea.
#
#This planet has - or rather had - a problem, which was this:  most of the
#people on it were unhappy for pretty much of the time.  Many solutions
#were suggested for this problem, but most of these were largely concerned
#with the movements of small green pieces of paper, which is odd because on
#the whole it wasn't the small pieces of paper that were unhappy.
#"""
#    self.book.words = [0, 9, 14, 17, 19, 28, 31, 39,
#                       50, 58, 64, 67, 70, 
#                       0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#                       0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#                       0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#                       0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#                       0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#                       0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#                       0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#                       0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#                       ]

    #i = 0
    #text = self.book.text
    #done = 0
    #while not done:
    #  i = string.find(text[i:], " ")
    #  if i >= 0:
    #    i = i+1
    #    self.book.words.append(i)
    #    i = i+1
    #    self.book.length = self.book.length + 1
    #  else:
    #    done = 1

#    self.book.types = ["", "", "", "", "", "", "", "",
#                      "", "", "", "", "", "", "", "", "", "",
#                      "", "", "", "", "", "", "", "", "", "", "",
#                      "", "", "", "", "", "", "", "\n",
#                      "", "", "", "", "", "", "", "", "", "", "", ",", "", "", "", ":", "", "", "",
#                      "", "", "", "", "", "", "", "", "", "", "", ".", "", "",
#                      "", "", "", "", "", ",", "", "", "", "", "", "", "",
#                      "", "", "", "", "", "", "", "", "", ",", "", "", "", "", "",
#                      "", "", "", "", "", "", "", "", "", "", "", "", "\n", ]
#    self.book.length = len(self.book.types)
    
#    self.words = [
#      ("Orbiting",""), ("this",""), ("at",""), ("a",""),
#      ("distance",""), ("of",""), ("roughly",""), ("ninety-two",""),
#      ("million",""), ("miles",""), ("is",""), ("an",""),
#      ("utterly",""), ("insignificant",""), ("little",""),
#      ("blue",""), ("green",""), ("planet",""), ("whose",""),
#      ("ape-descended",""), ("life",""), ("forms",""), ("are",""),
#      ("so",""), ("amazingly",""), ("primitive",""), ("that",""),
#      ("they",""), ("still",""), ("think",""), ("digital",""),
#      ("watches",""), ("are",""), ("a",""), ("pretty",""),
#      ("neat",""), ("idea.",""), ("","\n"), ("This",""),
#      ("planet",""), ("has",""), ("-",""), ("or",""), ("rather",""),
#      ("had",""), ("-",""), ("a",""), ("problem,",""), ("",","),
#      ("which",""), ("was",""), ("this:",""), ("",":"), ("most",""),
#      ("of",""), ("the",""), ("people",""), ("on",""), ("it",""),
#      ("were",""), ("unhappy",""), ("for",""), ("pretty",""),
#      ("much",""), ("of",""), ("the",""), ("time.",""), ("","."),
#      ("Many",""), ("solutions",""), ("were",""), ("suggested",""),
#      ("for",""), ("this",""), ("problem,",""), ("",","), ("but",""),
#      ("most",""), ("of",""), ("these",""), ("were",""),
#      ("largely",""), ("concerned",""), ("with",""), ("the",""),
#      ("movements",""), ("of",""), ("small",""), ("green",""),
#      ("pieces",""), ("of",""), ("paper,",""), ("",","), ("which",""),
#      ("is",""), ("odd",""), ("because",""), ("on",""), ("the",""),
#      ("whole",""), ("it",""), ("wasn't",""), ("the",""),
#      ("small",""), ("green",""), ("pieces",""), ("of",""),
#      ("paper",""), ("that",""), ("were",""), ("unhappy.",""),
#      ("","\n"), ("","c"),
#      ]
    


######################################################################
#
def run(args):
  cfg = config()
  cfg.load()

  testwords = 0
  infiles = []

  for arg in args:
    if arg == "-n"  or  arg == "-e"  or  arg == "--enter-pbreak":
      cfg.options["enter is paragraph break"] = 1

    else:
      infiles.append(arg)
    
  if len(args) <= 0:
    testwords = 1

  if infiles == []:
    v = viewer(cfg)
    v.make_test_words()
    v.run()
  else:
    for filename in infiles:
      v = viewer(cfg)
      v.load_plaintext(filename)
      v.run()
    
    


######################################################################
#
if __name__ == "__main__":
  import sys
  run(sys.argv[1:])
